﻿using System.Reflection;
using System.Reflection.Emit;

namespace Fipscode.CodeGenerator
{
    public static class Generator
    {
        public static AssemblyBuilder NewAssembly(string assemblyName, AssemblyBuilderAccess access = AssemblyBuilderAccess.Run)
        {
            AssemblyName aname = new AssemblyName(assemblyName);
            AssemblyBuilder abuilder = AssemblyBuilder.DefineDynamicAssembly(aname, access);
            return abuilder;
        }

        public static ModuleBuilder NewModule(this AssemblyBuilder abuilder, string moduleName = null, string dllName = null)
        {
            var modname = moduleName ?? abuilder.GetName().ToString();
            var libname = dllName ?? $"{modname}.dll";
            ModuleBuilder mbuilder;
            //try {
            //    mbuilder = abuilder.DefineDynamicModule(modname, libname);
            //} catch (NotSupportedException ex) {
                mbuilder = abuilder.DefineDynamicModule(modname);
            //}
            return mbuilder;
        }

        public static TypeBuilder NewType(this ModuleBuilder mbuilder, string className = null)
        {
            var typename = className ?? mbuilder.Name;
            TypeBuilder tbuilder = mbuilder.DefineType(typename, TypeAttributes.Public);
            return tbuilder;
        }

        public static TypeBuilder NewType(this ModuleBuilder mbuilder, string className = null, params string[] genericParameters)
        {
            var typename = className ?? mbuilder.Name;
            TypeBuilder tbuilder = mbuilder.DefineType(typename, TypeAttributes.Public);
            GenericTypeParameterBuilder[] genBuilders = tbuilder.DefineGenericParameters(genericParameters);
            foreach (GenericTypeParameterBuilder genBuilder in genBuilders) {
                genBuilder.SetGenericParameterAttributes(
                        GenericParameterAttributes.ReferenceTypeConstraint | GenericParameterAttributes.DefaultConstructorConstraint
                );
            }
            return tbuilder;
        }

        public static MethodBuilder NewMethod(this TypeBuilder tbuilder, string methodName)
        {
            MethodBuilder mbuilder = tbuilder.DefineMethod(methodName, MethodAttributes.Public | MethodAttributes.HideBySig);
            return mbuilder;
        }

        public static MethodBuilder NewMethod(this TypeBuilder tbuilder, string methodName, System.Type returnType, params System.Type[] parameterTypes)
        {
            MethodBuilder mbuilder = tbuilder.DefineMethod(
                    methodName,
                    MethodAttributes.Public | MethodAttributes.HideBySig,
                    CallingConventions.HasThis,
                    returnType,
                    parameterTypes
            );
            return mbuilder;
        }

        public static MethodBuilder NewMethod(this TypeBuilder tbuilder, string methodName, System.Type returnType, string[] genericParameters, params System.Type[] parameterTypes)
        {
            MethodBuilder mbuilder = tbuilder.DefineMethod(
                    methodName,
                    MethodAttributes.Public | MethodAttributes.HideBySig,
                    CallingConventions.HasThis,
                    returnType,
                    parameterTypes
            );
            GenericTypeParameterBuilder[] genBuilders = mbuilder.DefineGenericParameters(genericParameters);
            foreach (GenericTypeParameterBuilder genBuilder in genBuilders) {
                genBuilder.SetGenericParameterAttributes(
                        GenericParameterAttributes.ReferenceTypeConstraint | GenericParameterAttributes.DefaultConstructorConstraint
                );
            }
            return mbuilder;
        }
    }
}
