﻿using System;
using System.CodeDom.Compiler;
using System.Reflection;
using System.Reflection.Emit;
using System.Text;
using Microsoft.CSharp;
using s = System;
using sr = System.Reflection;

namespace Fipscode.CodeGenerator
{
    public class Program
    {
        static string _name;
        static string[] _variables;
        StringBuilder _sb;

        static Program()
        {
            _name = "Foobar";
            _variables = new[] { "foo", "bar" };
        }

        public Program(string foo, string bar)
        {
            _sb = new StringBuilder();
            _sb.Append(foo);
            _sb.Append(": ");
            _sb.Append(bar);
        }

        public string Format()
        {
            return _sb.ToString();
        }

        static void Main(string[] args)
        {
            CompileSource(args);
        }

        static void ByteCodeEmitting(string[] args)
        {
            var tb = Generator
                .NewAssembly("MyAssembly")
                //.NewAssembly("MyAssembly", AssemblyBuilderAccess.RunAndSave)
                .NewModule()
                .NewType("MyClass");
            var mb = tb
                .NewMethod("Sum", typeof(Single), new[] { typeof(Int32), typeof(Int32) });
            ILGenerator generator = mb.GetILGenerator();
            generator.DeclareLocal(typeof(Single));
            generator.Emit(OpCodes.Ldarg_1);
            generator.Emit(OpCodes.Ldarg_2);
            generator.Emit(OpCodes.Add_Ovf);
            generator.Emit(OpCodes.Conv_R4);
            generator.Emit(OpCodes.Stloc_0);
            generator.Emit(OpCodes.Ldloc_0);
            generator.Emit(OpCodes.Ret);
            var typ = tb.CreateTypeInfo().AsType();
            var inst = Activator.CreateInstance(typ);
            Console.WriteLine(typ.GetMethod("Sum").Invoke(inst, new object[] { 5, 10 }));
            //asm.Save(@"MyAssembly.dll");
            Console.ReadKey();
        }

        static void CompileSource(string[] args)
        {
            string code = @"
                using System;
                
                namespace DynamicallyCompiled
                {
                    public class Program
                    {
                        public static void Main()
                        {
                            Console.WriteLine(""Hello, world!"");
                        }
                    }
                }";
            CSharpCodeProvider provider = new CSharpCodeProvider();
            CompilerParameters parameters = new CompilerParameters();
            // true  - memory generation
            // false - external file generation
            parameters.GenerateInMemory = true;
            // true  - exe file generation
            // false - dll file generation
            parameters.GenerateExecutable = true;
            // optional: reference a library
            //parameters.ReferencedAssemblies.Add("Some.Library.dll");
            //var netstandard = sr.Assembly.Load("netstandard, Version=2.0.0.0, Culture=neutral, PublicKeyToken=cc7b13ffcd2ddd51");
            //parameters.ReferencedAssemblies.Add(netstandard.Location);
            CompilerResults results = provider.CompileAssemblyFromSource(parameters, code);
            if (results.Errors.HasErrors) {
                StringBuilder sb = new StringBuilder();
                foreach (CompilerError error in results.Errors) {
                    sb.AppendLine(String.Format("Error ({0}): {1}", error.ErrorNumber, error.ErrorText));
                }
                throw new InvalidOperationException(sb.ToString());
            }
            sr.Assembly assembly = results.CompiledAssembly;
            s.Type program = assembly.GetType("DynamicallyCompiled.Program");
            MethodInfo main = program.GetMethod("Main");
            main.Invoke(null, null);
            Console.ReadKey();
        }

        static string BuildString(params string[] parameters)
        {
            var sb = new StringBuilder();
            sb.Append("Text:\n");
            sb.Append(parameters[42]);
            sb.Append(": ");
            sb.Append(parameters[1]);
            return sb.ToString();
        }
    }
}
